"""Display catalog of commands user uses often."""

import platform

from commandsheet.cli import options
from commandsheet.cli import argparser
from commandsheet.const import XDG_CONFIG_PATH
from commandsheet.config import parse_config
from commandsheet.config import config_exists
from commandsheet.config import is_valid_config_file
from commandsheet.config import config_file_path_exists
from commandsheet.compatibility import compatible_os
from commandsheet.errors import not_a_valid_config_file
from commandsheet.errors import no_config_file_found
from commandsheet.errors import no_config_file_sections_found
from commandsheet.errors import no_config_file_path_exists
from commandsheet.errors import no_compatible_os
from commandsheet.output import header
from commandsheet.output import display_commandsheet


def main():
    parser = argparser()
    options(parser)
    args = parser.parse_args()

    operating_system = platform.system()
    if not compatible_os(operating_system):
        no_compatible_os(parser, operating_system)

    fillchar = args.fillchar
    section_numbers = args.section_numbers
    config_file = args.config_file if args.config_file else XDG_CONFIG_PATH

    if not config_file_path_exists(config_file):
        no_config_file_path_exists(parser, config_file)
    if not is_valid_config_file(config_file):
        not_a_valid_config_file(parser, file=config_file)
    if not config_exists(config_file):
        no_config_file_found(parser)

    commandsheet = parse_config(config_file)
    if not commandsheet:
        no_config_file_sections_found(parser)

    header()
    display_commandsheet(
        commandsheet, fillchar=fillchar, section_numbers=section_numbers,
    )


if __name__ == '__main__':
    main()



"""Code to validate if OS is compatible."""

COMPATIBLE_OS = (
    'linux',
    'darwin',
)


def compatible_os(name):
    if name.strip().lower() not in COMPATIBLE_OS:
        return False
    return True



from configparser import ConfigParser
from contextlib import contextmanager
from pathlib import Path
from os import chdir

from commandsheet.config import config_file_path_exists
from commandsheet.config import is_valid_config_file
from commandsheet.config import config_empty
from commandsheet.config import config_exists
from commandsheet.config import parse_config
from commandsheet.config import Section

import pytest


@contextmanager
def inside_dir(path):
    old_path = Path.cwd()
    try:
        chdir(path)
        yield
    finally:
        chdir(str(old_path))


def test_section():
    section = Section(name='section', contents=('cmd', 'desc'))
    assert section.name == 'section'
    assert section.contents == ('cmd', 'desc')


def test_config_empty():
    empty_config = ConfigParser()
    not_empty_config = ConfigParser()
    with inside_dir('example'):
        assert config_empty(empty_config)
        not_empty_config.read('commandsheet.ini')
        assert not config_empty(not_empty_config)


def test_config_exists():
    exist_file = 'commandsheet.ini'
    no_exist_file = 'no_exist_commandsheet.ini'
    with inside_dir('example'):
        assert config_exists(exist_file)
        assert not config_exists(no_exist_file)


def test_is_valid_config_file():
    assert is_valid_config_file('commandsheet.ini')
    assert is_valid_config_file('also-valid.ini')
    assert not is_valid_config_file('config.invalid')


def test_config_file_path_exists():
    assert not config_file_path_exists('invalid/path/to/file')
    with inside_dir('example'):
        assert config_file_path_exists('commandsheet.ini')


def test_parse_config():
    with inside_dir('example'):
        not_empty_commandsheet = parse_config('commandsheet.ini')
        assert not_empty_commandsheet != []
        empty_commandsheet = parse_config('')
        assert empty_commandsheet == []



"""Tests for testing CLI errors."""

from pathlib import Path

from commandsheet.cli import CustomArgumentParser
from commandsheet.errors import no_config_file_found
from commandsheet.errors import no_config_file_sections_found
from commandsheet.errors import no_config_file_path_exists
from commandsheet.errors import not_a_valid_config_file
from commandsheet.errors import no_compatible_os

import pytest


def test_no_config_file_found(capsys):
    parser = CustomArgumentParser()
    with pytest.raises(SystemExit):
        no_config_file_found(parser)

    out, err = capsys.readouterr()
    assert err == 'No `commandsheet.ini` file found :(\n'
    assert out == ''


def test_no_config_file_sections_found(capsys):
    parser = CustomArgumentParser()
    with pytest.raises(SystemExit):
        no_config_file_sections_found(parser)

    out, err = capsys.readouterr()
    assert err == 'No sections to display from commandsheet.ini :(\n'
    assert out == ''


def test_no_config_file_path_exists(capsys):
    parser = CustomArgumentParser()
    file = '/invalid/path/to/file.ini'
    with pytest.raises(SystemExit):
        no_config_file_path_exists(parser, file=file)

    out, err = capsys.readouterr()
    assert err == f'Config file `{file}` does not exist\n'
    assert out == ''


def test_not_a_valid_config_file(capsys):
    parser = CustomArgumentParser()
    file = '/path/to/invalid/configfile.invalid'
    with pytest.raises(SystemExit):
        not_a_valid_config_file(parser, file=file)

    out, err = capsys.readouterr()
    name = Path(file).name
    assert err == f'File `{name}` is not of valid config file format\n'
    assert out == ''


def test_no_compatible_os(capsys):
    parser = CustomArgumentParser()
    os = 'Windows'

    with pytest.raises(SystemExit):
        no_compatible_os(parser, os)

    out, err = capsys.readouterr()
    assert err == f'`{os}` is not an OS that is supported :(\n'
    assert out == ''



from textwrap import dedent

from commandsheet.config import Section
from commandsheet.output import header
from commandsheet.output import display_commandsheet
from commandsheet.output import format_section_heading
from commandsheet.output import format_section_content

import pytest


def test_header(capsys):
    """
    ████████████████████████████████████████████████████████████████████████████
    █─▄▄▄─█─▄▄─█▄─▀█▀─▄█▄─▀█▀─▄██▀▄─██▄─▀█▄─▄█▄─▄▄▀█─▄▄▄▄█─█─█▄─▄▄─█▄─▄▄─█─▄─▄─█
    █─███▀█─██─██─█▄█─███─█▄█─███─▀─███─█▄▀─███─██─█▄▄▄▄─█─▄─██─▄█▀██─▄█▀███─███
    ▀▄▄▄▄▄▀▄▄▄▄▀▄▄▄▀▄▄▄▀▄▄▄▀▄▄▄▀▄▄▀▄▄▀▄▄▄▀▀▄▄▀▄▄▄▄▀▀▄▄▄▄▄▀▄▀▄▀▄▄▄▄▄▀▄▄▄▄▄▀▀▄▄▄▀▀
    """
    header()
    out, err = capsys.readouterr()
    assert out == dedent(test_header.__doc__) + '\n'
    assert err == ''


def test_format_section_heading():
    text = 'heading'
    index = 1
    assert format_section_heading(text, index=index) == '1. heading'
    assert format_section_heading(text, index=None) == 'heading'


def test_format_section_content_with_short_description():
    cmd = 'ls --long'
    dsc = 'List in a long format'
    fillchar = '.'
    indent = len(cmd) + 5
    max_width = 85

    result = format_section_content(
        cmd,
        dsc,
        indent=indent,
        fillchar=fillchar,
        max_width=max_width
    )

    expected = 'ls --long..... List in a long format'
    assert result == expected


def test_format_section_content_with_long_description():
    cmd = 'ls --long'
    dsc = 'List in a long format. You can also use the -l option.'
    fillchar = '.'
    indent = len(cmd) + 5

    # We don't need to actually write a longer description
    # to test against a 'longer' description (we could if
    # we'd like to, but that is unnecessary). Instead, we
    # can simply reduce the value of the ``max_width``
    # parameter.
    max_width = 40

    result = format_section_content(
        cmd,
        dsc,
        indent=indent,
        fillchar=fillchar,
        max_width=max_width
    )

    expected = (
        'ls --long..... List in a long format. You can also use\n'
        '               the -l option.'
    )
    assert result == expected


def test_display_commandsheet(capsys):
    content = (('cmd', 'desc'),)
    alpha = Section(name='alpha', contents=content)
    bravo = Section(name='bravo', contents=content)

    # This basically represents the commandsheet as if
    # it would be read from a file and parsed. This way,
    # we don't have to actually have a config-file that
    # we have to read, but we can simulate one, instead.
    sections = [alpha, bravo]

    # We're not using the ``section_numbers`` parameter here
    display_commandsheet(
        sections,
        fillchar='.',
        section_numbers=False,
        max_width=40,
        indent=8
    )

    expected = (
        '╭─ alpha ───────╮\n'
        '│ cmd..... desc │\n'
        '╰───────────────╯\n'
        '╭─ bravo ───────╮\n'
        '│ cmd..... desc │\n'
        '╰───────────────╯\n'
    )

    out, err = capsys.readouterr()
    assert out == expected
    assert err == ''

    # We are using the ``section_numbers`` parameter here
    display_commandsheet(
        sections,
        fillchar='.',
        section_numbers=True,
        max_width=40,
        indent=8
    )

    expected = (
        '╭─ 1. alpha ────╮\n'
        '│ cmd..... desc │\n'
        '╰───────────────╯\n'
        '╭─ 2. bravo ────╮\n'
        '│ cmd..... desc │\n'
        '╰───────────────╯\n'
    )

    out, err = capsys.readouterr()
    assert out == expected
    assert err == ''



"""Code for displaying the command sheet -catalog."""

from textwrap import dedent
from textwrap import wrap


def header():
    """
    ████████████████████████████████████████████████████████████████████████████
    █─▄▄▄─█─▄▄─█▄─▀█▀─▄█▄─▀█▀─▄██▀▄─██▄─▀█▄─▄█▄─▄▄▀█─▄▄▄▄█─█─█▄─▄▄─█▄─▄▄─█─▄─▄─█
    █─███▀█─██─██─█▄█─███─█▄█─███─▀─███─█▄▀─███─██─█▄▄▄▄─█─▄─██─▄█▀██─▄█▀███─███
    ▀▄▄▄▄▄▀▄▄▄▄▀▄▄▄▀▄▄▄▀▄▄▄▀▄▄▄▀▄▄▀▄▄▀▄▄▄▀▀▄▄▀▄▄▄▄▀▀▄▄▄▄▄▀▄▀▄▀▄▄▄▄▄▀▄▄▄▄▄▀▀▄▄▄▀▀
    """
    print(dedent(header.__doc__))


def format_section_heading(heading, index):
    if index is not None:
        return f'{index}. {heading}'
    return heading


# Tries to achieve the following:
#
#    <section command>............... <section's way too long description
#                                      string here, so it is split like
#                                      this, like the argparse does with
#                                      the help messages> :)
#
# Instead of:
#
#    <section command>............... <section's way too long description
#    string here, which is not split like the example above, and looks kinda
#    ugly and not readable, especially when more commands starts to pile up> :(
#
def format_section_content(cmd, desc, indent, fillchar, max_width):
    parts = wrap(desc, width=max_width)
    indented = []

    # Construct the description
    for index, sentence in enumerate(parts):
        if index == 0:
            indented.append(sentence)
        else:
            indented.append(' ' + (' ' * indent) + sentence)
    wrapped = '\n'.join(indented)
    description = wrapped

    result = f'{cmd:{fillchar}<{indent}} {description}'
    return result


def display_commandsheet(
    commandsheet,
    *,
    fillchar,
    section_numbers,
    max_width=40,
    indent=50
):
    from rich import print as rich_print
    from rich.panel import Panel

    for idx, section in enumerate(commandsheet, start=1):
        # Format section heading
        section_heading = format_section_heading(
            section.name, index=idx if section_numbers else None,
        )

        # Format section contents
        section_contents = []
        for cmd, desc in section.contents:
            line = format_section_content(
                cmd,
                desc,
                indent=indent,
                fillchar=fillchar,
                max_width=max_width
            )
            section_contents.append(line)

        # Print section with nice border
        content = '\n'.join(section_contents)
        panel = Panel(
            content,
            title=section_heading,
            title_align='left',
            expand=False
        )
        rich_print(panel)



"""Code for parsing the user config."""

from attrs import define
from attrs import field
from pathlib import Path
from configparser import ConfigParser

CONFIG_FILE_FORMATS = {'.ini'}


@define
class Section:
    name = field()
    contents = field()


def config_empty(config):
    return not config.sections()


def config_exists(path):
    return Path(path).expanduser().exists()


def is_valid_config_file(path):
    return Path(path).expanduser().suffix in CONFIG_FILE_FORMATS


def config_file_path_exists(path):
    """Verifies that, when `commandsheet -c <config>`, that <config> exists."""
    return config_exists(path)


def parse_config(path):
    config = ConfigParser()
    config.read(Path(path).expanduser())

    if config_empty(config):
        return []

    sections_raw = [
        section for section in config.values()
        # In our case, we don't care about the DEFAULT
        # section configparser creates to the config
        # object by default, so we'll just leave it out
        if not section.name == 'DEFAULT'
    ]

    sections_pretty = []
    for section in sections_raw:
        content = []
        for cmd, desc in section.items():
            content.append((cmd, desc))
        sections_pretty.append(Section(name=section.name, contents=content))

    return sections_pretty



import sys
from commandsheet.cli import CustomArgumentParser
from commandsheet.cli import argparser
from commandsheet.cli import options
import pytest


class TestCustomArgumentParser:
    def test_print_usage(self, capsys):
        parser = CustomArgumentParser()

        parser.print_usage()
        out, err = capsys.readouterr()
        assert out.startswith('usage: ')
        assert err == ''

        parser.print_usage(file=sys.stderr)
        out, err = capsys.readouterr()
        assert out == ''
        assert err.startswith('usage: ')

    def test_print_help(self, capsys):
        parser = CustomArgumentParser()

        parser.print_help()
        out, err = capsys.readouterr()
        assert out.startswith('usage: ')
        assert err == ''

        parser.print_help(file=sys.stderr)
        out, err = capsys.readouterr()
        assert out == ''
        assert err.startswith('usage: ')

    def test_parse_args(self, capsys):
        prog = 'test'
        command = 'unknown-command'
        parser = CustomArgumentParser(prog=prog)

        with pytest.raises(SystemExit):
            parser.parse_args((command,))
        out, err = capsys.readouterr()
        assert out == ''
        assert err.split('\n')[0] == f'usage: {prog} [-h]'
        assert err.split('\n')[1] == f'{prog}: error: unrecognized arguments: {command}'

        args = parser.parse_args()
        out, err = capsys.readouterr()
        assert out == ''
        assert err == ''

    def test_exit(self, capsys):
        parser = CustomArgumentParser()

        with pytest.raises(SystemExit):
            parser.exit(status=0, message='Exit message', newline=True)
        out, err = capsys.readouterr()
        assert out == ''
        assert err == 'Exit message\n'

        with pytest.raises(SystemExit):
            parser.exit(status=0, message='Exit message', newline=False)
        out, err = capsys.readouterr()
        assert out == ''
        assert err == 'Exit message'

        with pytest.raises(SystemExit):
            parser.exit(status=0)
        out, err = capsys.readouterr()
        assert out == ''
        assert err == ''

    def test_error(self, capsys):
        prog = 'test'
        msg = 'this is an error message'
        parser = CustomArgumentParser(prog=prog)

        with pytest.raises(SystemExit):
            parser.error(message=msg)
        out, err = capsys.readouterr()
        assert out == ''
        assert err.split('\n')[0] == f'usage: {prog} [-h]'
        assert err.split('\n')[1] == f'{prog}: error: {msg}'


def test_options():
    parser = CustomArgumentParser(add_help=False)
    options(parser)
    args = parser.parse_args()

    assert hasattr(args, 'section_numbers')
    assert hasattr(args, 'fillchar')
    assert hasattr(args, 'config_file')


def test_argparser():
    parser = argparser()
    assert isinstance(parser, CustomArgumentParser)
    assert parser.prog == 'commandsheet'
    assert parser.description == 'Display catalog of commands user uses often.'
    assert parser.epilog is None
    assert parser.usage is None
    assert parser.add_help == False
    assert parser.allow_abbrev == False



from setuptools import setup


def readme() -> str:
    with open("README.md", "r", encoding="utf-8") as f:
        return f.read()


setup(
    # Name of your project. When you publish this
    # package to PyPI, this name will be registered for you.
    name='commandsheet',  # Required

    # Version?
    version='0.1.0',  # Required

    # What does your project do?
    description="Display catalog of commands user uses often.",  # Optional

    # Longer description that users will see when
    # they visit your project at PyPI.
    long_description=readme(),  # Optional

    # Denotes that long description is in Markdown.
    # Valid values are: text/plain, text/x-rst or text/markdown.
    # Optional if 'long_description' is written in rst, otherwise
    # required (for plain-text and Markdown).
    long_description_content_type='text/markdown',  # Optional

    # Who owns this project?
    author='Niklas Larsson',  # Optional

    # Project owner's email
    author_email='',  # Optional

    # More info at: https://pypi.org/classifiers/
    classifiers=[  # Optional
        'License :: OSI Approved :: MIT License',
    ],

    # What does your project relate to?
    keywords='',  # Optional

    # Is your project larger than just few files?
    packages=['commandsheet'],  # Required

    # Which Python versions are supported?
    # e.g. 'pip install' will check this and refuse to install
    # the project if the version doesn't match.
    python_requires='>=3.7',  # Optional

    # Any dependencies?
    install_requires=['attrs', 'rich'],  # Optional

    # Need to install, for example, man-pages that your project has?
    package_data={
        'example': ['commandsheet.ini'],
    },

    # Any executable scripts?
    entry_points={  # Optional
        "console_scripts": [
            'commandsheet = commandsheet.commandsheet:main',
        ]
    },

    # More info at: https://setuptools.pypa.io/en/latest/userguide/datafiles.html
    include_package_data=True,  # Optional

    # More info at: https://setuptools.pypa.io/en/latest/userguide/miscellaneous.html
    zip_safe=False,  # Optional

    # Additional URLs that are relevant to your project?
    project_urls={  # Optional
        'Source': 'https://github.com/nikkelarsson/commandsheet',
    }
)



from commandsheet.compatibility import compatible_os


def test_compatible_os_with_compatible_os():
    oses = ('Linux', 'Darwin',)
    for os in oses:
        assert compatible_os(os)


def test_compatible_os_with_incompatible_os():
    oses = ('Windows', 'Other invalid OS',)
    for os in oses:
        assert not compatible_os(os)



"""CLI errors."""

from pathlib import Path
from commandsheet.locale import _


def no_config_file_found(parser):
    msg = _('No `commandsheet.ini` file found :(')
    parser.exit(status=1, message=msg, newline=True)


def no_config_file_sections_found(parser):
    msg = _('No sections to display from commandsheet.ini :(')
    parser.exit(status=1, message=msg, newline=True)


def no_config_file_path_exists(parser, file):
    path = Path(file).expanduser()
    msg = _('Config file `%(file)s` does not exist')
    parser.exit(status=1, message=msg % {'file': str(path)}, newline=True)


def not_a_valid_config_file(parser, file):
    path = Path(file).expanduser()
    filename = Path(path).expanduser().name
    msg = _('File `%(file)s` is not of valid config file format')
    parser.exit(status=1, message=msg % {'file': filename}, newline=True)


def no_compatible_os(parser, os_name):
    msg = _('`%(os)s` is not an OS that is supported :(')
    parser.exit(status=1, message=msg % {'os': os_name}, newline=True)



"""Code for the command-line interface."""

import argparse
import sys
from commandsheet.locale import _


class CustomArgumentParser(argparse.ArgumentParser):
    def print_usage(self, file=None):
        file = sys.stdout if file is None else sys.stderr
        output = self.format_usage().replace('usage: ', _('usage: '), 1)
        print(output, file=file, end='')

    def print_help(self, file=None):
        file = sys.stdout if file is None else sys.stderr
        output = self.format_help().replace('usage: ', _('usage: '), 1)
        print(output, file=file, end='')

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        if argv:
            unrecognized = {'args': ' '.join(argv)}
            msg = _('unrecognized arguments: %(args)s' % unrecognized)
            self.error(msg)
        return args

    def exit(self, status=0, message=None, newline=False):
        if message is not None:
            end_char = '' if not newline else '\n'
            print(_(message), file=sys.stderr, end=end_char)
        sys.exit(status)

    def error(self, message):
        self.print_usage(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(
            status=2,
            newline=True,
            message=_('%(prog)s: error: %(message)s' % args)
        )


def options(parser):
    opts = parser.add_argument_group(_('options'))
    opts.add_argument(
        '-s',
        '--section-numbers',
        action='store_true',
        help=_("Print a section number in front of each section title")
    )
    opts.add_argument(
        '-f',
        '--fillchar',
        action='store',
        metavar=_('CHAR'),
        default='.',
        type=str,
        help=_(
            "Use %(metavar)s to fill the gap between "
            "the command and its explanation "
            "(default: `%(default)s`)"
        )
    )
    opts.add_argument(
        '-c',
        '--config-file',
        action='store',
        metavar=_('CONFIG'),
        help=_("Use %(metavar)s as the config file")
    )
    opts.add_argument(
        '-h',
        '--help',
        action='help',
        help=_("Show this message")
    )
    opts.add_argument(
        '-V',
        '--version',
        action='version',
        version='%(prog)s 0.1.0',
        help=_("Show version")
    )


def argparser():
    return CustomArgumentParser(
        prog='commandsheet',
        description=_("Display catalog of commands user uses often."),
        add_help=False,
        allow_abbrev=False
    )



from pathlib import Path
from contextlib import contextmanager
from commandsheet.const import XDG_CONFIG_PATH
import subprocess
import pytest


SUCCESS = 0


def test_runas_entrypoint():
    cmd = ('commandsheet', '--help',)
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == SUCCESS


def test_runas_module():
    cmd = ('python3', '-m', 'commandsheet', '--help',)
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == SUCCESS


def test_version_option():
    cmd = ('commandsheet', '--version',)
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == SUCCESS
    assert result.stdout == 'commandsheet 0.1.0\n'


@contextmanager
def no_config_file(path):
    """Temporarily renames the config file to
    make commandsheet think it does not exist,
    enabling us to simulate the scenario where
    the config file wouldn't actually exist.
    """
    exists = Path(path).expanduser().exists()
    original = str(Path(path).expanduser().parent) + '/' + 'commandsheet.ini'
    temp = str(Path(path).expanduser().parent) + '/' + '_commandsheet.ini'
    try:
        if exists:
            Path(original).replace(temp)
        yield
    finally:
        if exists:
            Path(temp).replace(original)


def test_invocation_without_existing_config_file():
    with no_config_file(XDG_CONFIG_PATH):
        cmd = ('commandsheet')
        result = subprocess.run(cmd, capture_output=True, text=True)
        assert not result.returncode == SUCCESS


def test_config_file_option():
    cmd = ('commandsheet', '--config-file', 'example/commandsheet.ini')
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == SUCCESS

    cmd = ('commandsheet', '--config-file', 'invalid/path/to/file.invalid')
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert not result.returncode == SUCCESS

    cmd = ('commandsheet', '--config-file', 'setup.py')
    result = subprocess.run(cmd, capture_output=True, text=True)
    assert not result.returncode == SUCCESS
